import re
from pathlib import Path

from chardet.universaldetector import UniversalDetector

__char_detector = UniversalDetector()


def detect_encoding(filename):
    __char_detector.reset()
    try:
        with open(filename, 'rb') as fp:
            for line in fp.readlines():
                __char_detector.feed(line)
                if __char_detector.done:
                    break
            result = __char_detector.result
            return result['encoding']
    except FileNotFoundError:
        return 'utf-8'


def normalize_path(path: Path, anchor=Path('.')) -> Path:
    path = Path(path)
    anchor = Path(anchor)
    if path.is_absolute():
        return path.absolute()

    t = str(path)
    if t.startswith('~'):
        return Path(t.replace('~', str(Path.home()))).absolute()

    return anchor.absolute() / path


def quote_path(path, quote_char='"') -> str:
    path = str(path)
    return f'{quote_char}{path}{quote_char}' if ' ' in path else path


def limit_number(x, left, right):
    _min = min(left, right)
    _max = max(left, right)
    if x < _min:
        return _min
    if x > _max:
        return _max
    return x


def split_at_unquoted_spaces(line: str):
    result = []
    buffer = []
    symbols = {'"', "'", '(', ')', '[', ']', '{', '}'}
    matching = {')': '(', ']': '[', '}': '{'}
    stack = []
    in_quotes = False
    quote_char = ''

    for char in str(line):
        if char in symbols:
            if in_quotes:
                if char == quote_char:
                    in_quotes = False
            else:
                if char == '"' or char == "'":
                    in_quotes = True
                    quote_char = char
                elif char in matching and stack and stack[-1] == matching[char]:
                    stack.pop()
                else:
                    stack.append(char)
        if char == ' ' and not stack and not in_quotes:
            result.append(''.join(buffer))
            buffer = []
        else:
            buffer.append(char)

    # 如果缓冲区还有字符，则将它们加入到结果列表中。
    if buffer:
        result.append(''.join(buffer))

    return result


def unquote_text(s: str):
    """删除字符串前后的成对引号以及多余的空格"""
    pattern = r'^("|\')((?:(?=(\\?))\3.)*?)\1$'
    text = s.strip()
    match = re.match(pattern, text)
    if match:
        text = match.group(2)
    return text.strip()


def quote_text(s: str, quote='"'):
    """阔起字符串，默认使用引号。如果给定的quote包含多个字符，则会使用前两个分别阔起，例如使用括号"""
    left = quote[0]
    right = quote[1] if len(quote) > 1 else left
    return left + str(s) + right



import asyncio
import json

from ffmpeg import Progress
from ffmpeg.asyncio import FFmpeg
from ffmpeg.errors import FFmpegError

from cx_core import DataPackage
from .env import env
from .mission import Mission


class Transcoder:
    def __init__(self, mission: Mission):
        self.task = None
        self.media_info = None
        self.mission = mission

    def __enter__(self):
        self.task = env.progress.add_task(description=f'{self.mission.source.name}...')

        ffprobe = FFmpeg(executable='ffprobe').input(self.mission.source, print_format='json', show_format=None)
        media = json.loads(asyncio.run(ffprobe.execute()))
        self.media_info = DataPackage(**media)

        duration = self.media_info.format.duration
        env.progress.update(self.task, total=float(duration) if duration else None)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.remove_task(self.task)
        self.task = None
        env.current_ffmpeg = None
        return False

    async def transcode(self):
        if not self.mission.check_output_writable():
            env.error(f'[red]任务的目标文件已存在且不可覆盖[/red]，自动跳过。')
            return

        for folder in self.mission.iter_target_folders():
            env.progress.update(self.task, description=f'检测目标目录: {folder}')
            if not folder.exists():
                env.print(f'目标目录 {folder} 不存在，自动创建')
                folder.mkdir(parents=True, exist_ok=True)

        ffmpeg = FFmpeg(self.mission.profile.general.ffmpeg)
        env.debug(f'创建 ffmpeg 对象: {self.mission.profile.general.ffmpeg}')

        env.progress.update(self.task, description=f'写入全局参数')
        for k, v in self.mission.general_options.iter_options():
            ffmpeg.option(k, v)
            env.debug(f'写入全局选项： [green]{k}[/green] : [yellow]{v}[/yellow]')

        for input_ in self.mission.inputs:
            env.progress.update(self.task, description=f'为 {input_.filename.name} 写入输入参数')
            ffmpeg.input(input_.filename.absolute(), input_.raw_data)
            env.debug(f'添加输入选项： {input_.raw_data}')

        for output_ in self.mission.outputs:
            env.progress.update(self.task, description=f'为 {output_.filename.name} 写入输出参数')
            ffmpeg.output(output_.filename.absolute(), output_.raw_data)
            env.debug(f'添加输出选项： {output_.raw_data}')

        env.debug(f'全部任务选项输入完毕， ffmpeg 对象构建完成')

        @ffmpeg.on('progress')
        def on_progress(progress: Progress):
            desc = '{0} [yellow]x{1:.2f}[/yellow]'.format(self.mission.source.name, progress.speed)
            curr = progress.time.seconds
            env.progress.update(self.task, description=desc, completed=curr)

        @ffmpeg.on('stderr')
        def on_stderr(line):
            env.debug(f'[grey]FFMPEG输出：[/grey] {line}')

        @ffmpeg.on('start')
        def on_start(arguments):
            env.debug(f'开始执行任务: {' '.join(arguments)}')

        @ffmpeg.on('completed')
        def on_completed():
            env.debug(f'[green]{self.mission.source.name}[/green] [yellow]执行完毕，顺利退出[/yellow]')

        @ffmpeg.on('terminated')
        def on_terminated():
            env.warning('[purple]FFMPEG 被终止[purple]')
            env.print('尝试移除未完成的目标文件...')
            for t in self.mission.outputs:
                f = t.filename
                f.unlink(missing_ok=True)
                env.debug(f'[red]已删除 {f.absolute()}[/red]')

        env.current_ffmpeg = ffmpeg
        await ffmpeg.execute()
        env.current_ffmpeg = None

    def run(self):
        try:
            asyncio.run(self.transcode())
        except FFmpegError as err:
            env.error(f'ffmpeg执行出错: [red]{err.message}[/red]')



class ProfileNoFoundError(FileNotFoundError):
    def __init__(self, target=None):
        super(ProfileNoFoundError, self).__init__()
        self._tgt = target

    def __str__(self):
        if self._tgt:
            return f'ffmpeg 配置文件 "{self._tgt}" 未找到，程序无法执行'
        return '未指定配置文件，程序无法运行'



import tomllib
from collections.abc import Iterable
from pathlib import Path

from cx_core import DataPackage
from cx_core.utils import normalize_path
from .env import env
from .exceptions import ProfileNoFoundError


class ProfileLoader:
    def __init__(self, filename):
        self.task_id = None
        self.filename = Path(filename)

    def __enter__(self):
        env.info('进入配置解析器环境')
        self.task_id = env.progress.add_task(description='初始化配置解析器', start=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.remove_task(self.task_id)
        self.task_id = None
        env.info('配置解析器已关闭')

    @staticmethod
    def __make_sure_list(a=None):
        if not a:
            return []
        if isinstance(a, str):
            return a.split(' ')
        if isinstance(a, Iterable):
            return [str(x) for x in a]
        return [x for x in str(a).split(' ')]

    def __check_data_package(self, package):
        env.progress.update(self.task_id, description='解析配置文件数据', completed=1, total=2)
        env.info('开始检查配置数据...')

        result = DataPackage(**package)
        env.print(f'开始解析配置文件：[cyan]{result.general.name}[/cyan]:[yellow]{result.general.description}[/yellow]')

        if not result.general.ffmpeg:
            result.general.ffmpeg = 'ffmpeg'
            env.print('未指定 ffmpeg ，将调用系统环境中的 ffmpeg')

        profile_folder = normalize_path('.' if not result.profile_path else result.profile_path.parent)
        result.profile_folder = profile_folder
        if not result.general.working_folder:
            env.debug('配置文件未指定工作目录')
            result.general.working_folder = profile_folder
        else:
            result.general.working_folder = normalize_path(result.general.working_folder, profile_folder)
        env.debug(f'general.working_folder 已设置为 {result.general.working_folder}')

        result.source.files.extend(ProfileLoader.__make_sure_list(env.args.sources))
        env.print(f'增加 {len(env.args.sources)} 个手动指定的来源文件')

        if not result.target.folder:
            env.debug('未设置目标文件夹')
            result.target.folder = Path('.')
        result.target.folder = normalize_path(result.target.folder, result.general.working_folder)
        env.debug(f'target.folder 已设置为 {result.target.folder}')

        if not result.target.keep_parent_level:
            env.debug('未设置目标父目录保留层级，将设为 0')
            result.target.keep_parent_level = 0

        env.debug(f'参数解析结果：{result}')

        return result

    def load(self):
        result = {}
        env.progress.start_task(self.task_id)
        env.progress.update(self.task_id, completed=0, total=2)
        task = env.progress.add_task(description=f'读取{self.filename}...')
        try:
            with env.progress.open(self.filename, 'rb', task_id=task) as fp:
                result.update(tomllib.load(fp))
            result['profile_path'] = Path(self.filename).absolute()
            return self.__check_data_package(result)
        except FileNotFoundError:
            raise ProfileNoFoundError(self.filename)
        finally:
            env.progress.remove_task(task)



from pathlib import Path

from cx_core.utils import quote_text
from .env import env
from .mission import Mission


class ScriptWriter:
    def __init__(self, target):
        self.target = Path(target).absolute()
        self.output = None
        self.planned_folders = set()

    def __enter__(self):
        self.output = open(self.target, 'w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.output.flush()
        self.output.close()
        return False

    @staticmethod
    def compile_cmd(mission: Mission):
        """将Mission编译为命令行"""
        line = [mission.profile.general.ffmpeg]
        line += mission.general_options.iter_arguments()
        for input_ in mission.inputs:
            line += input_.iter_arguments()
            line += ['-i', quote_text(input_.filename)]
        for output_ in mission.outputs:
            line += output_.iter_arguments()
            line += quote_text(output_.filename)
        return ' '.join([item for item in line if item is not None])

    def check_target_folders(self, mission):
        """检查潜在的目标文件夹，并写入新建文件夹的命令"""
        for folder in mission.iter_target_folders():
            if folder in self.planned_folders:
                continue
            self.output.write(f'mkdir -p {quote_text(folder.absolute())}\n')
            self.planned_folders.add(folder)
            env.info(f'计划创建目录： {folder}')

    def write_mission(self, mission: Mission):
        self.check_target_folders(mission)

        cmd = ScriptWriter.compile_cmd(mission)
        self.output.write(f'{cmd}\n')
        env.info(f'写入命令： [purple]{cmd}[/purple]')



from functools import cache
from pathlib import Path

from cx_core import DataPackage
from cx_core.tag_replacer import TagReplacer
from cx_core.utils import limit_number, split_at_unquoted_spaces, normalize_path
from .env import env
from .option_package import OptionPackage


class OptionParser:
    def __init__(self, options):
        self._options = options

    @staticmethod
    def _iter_list(elements: list):
        prev_token = None
        for t in elements:
            token = str(t).strip()
            if token.startswith('-'):
                if prev_token:
                    yield prev_token, None
                    prev_token = None
                else:
                    prev_token = token[1:]
            else:
                if prev_token:
                    yield prev_token, token
                    prev_token = None
                else:
                    env.debug(f'忽略无法识别的参数： {token}')
        if prev_token:
            yield prev_token, None

    @staticmethod
    def _iter_dict(elements: dict):
        for k, v in elements.items():
            yield k, v

    def __iter__(self):
        if isinstance(self._options, list):
            return self._iter_list(self._options)
        if isinstance(self._options, dict | DataPackage):
            return self._iter_dict(self._options)
        return self._iter_list(split_at_unquoted_spaces(str(self._options)))


class Mission:
    def __init__(self, source: Path, profile):
        self.source = Path(source)
        self.profile = profile
        self.tag_replacer = TagReplacer(keep_unknown_tags=False)
        self.tag_replacer \
            .install_data_source('source', self.__source_handler) \
            .install_data_source('target', self.__target_handler) \
            .install_data_source('source_parent', self.__source_parent_handler) \
            .install_data_source('target_parent', self.__target_parent_handler)

        self.general_options = self.__parse_general_table(self.profile.general)
        self.inputs: [OptionPackage] = [self.__parse_io_table(i) for i in self.profile.input]
        self.outputs: [OptionPackage] = [self.__parse_io_table(o) for o in self.profile.output]

    def __parse_general_table(self, general_table: DataPackage):
        result = OptionPackage()
        option_pairs = OptionParser(general_table.options)
        for k, v in option_pairs:
            value = self.tag_replacer(v)
            result.insert(k, value)

        if general_table.hardware_accelerate:
            result.insert('hwaccel', general_table.hardware_accelerate)

        overwrite_option = '-y' if general_table.overwrite_existed else '-n'
        result.insert(overwrite_option)
        return result

    def __parse_io_table(self, io_table: DataPackage) -> OptionPackage:
        filename = self.tag_replacer(io_table.filename)
        result = OptionPackage(filename=filename)
        opt_parser = OptionParser(io_table.options)
        for k, v in opt_parser:
            result.insert(k, self.tag_replacer(v))
        return result

    @cache
    def __source_handler(self, param=None):
        match param:
            case 'absolute':
                return str(self.source.absolute())
            case 'dot_suffix':
                return str(self.source.suffix)
            case 'suffix':
                return str(self.source.suffix)[1:]
            case 'parent':
                return str(self.source.parent)
            case 'parent_name':
                return str(self.source.parent.stem)
            case 'name':
                return str(self.source.name)
            case 'basename':
                return str(self.source.stem)
            case _:
                return str(self.source)

    @cache
    def __target_handler(self, param=None):
        match param:
            case 'absolute':
                return str(self.target.absolute())
            case 'dot_suffix':
                return str(self.target.suffix)
            case 'suffix':
                return str(self.target.suffix)[1:]
            case 'parent':
                return str(self.target.parent)
            case 'parent_name':
                return str(self.target.parent.stem)
            case 'name':
                return str(self.target.name)
            case 'basename':
                return str(self.target.stem)
            case _:
                return str(self.target)

    @cache
    def __source_parent_handler(self, param=None):
        level = int(param) if param else 1
        ps = self.source.parent.parts
        level = limit_number(level, 1, len(ps))
        selected_parts = ps[-1 * level:]
        return str(Path(*selected_parts))

    @cache
    def __target_parent_handler(self, param=None):
        level = int(param) if param else 1
        ps = self.target.parent.parts
        level = limit_number(level, 1, len(ps))
        selected_parts = ps[-1 * level:]
        return str(Path(*selected_parts))

    @property
    @cache
    def target(self) -> Path:
        target_folder = normalize_path(self.profile.target.folder, self.profile.general.working_folder)

        t_suffix = self.profile.target.suffix
        if not t_suffix.startswith('.'):
            t_suffix = '.' + t_suffix

        p_folder = Path()
        p_level = self.profile.target.keep_parent_level
        if p_level > 0:
            parents = self.source.parent.parts
            selected_parts = parents[-1 * p_level:]
            p_folder = Path(*selected_parts)
            env.debug(f'取用 {p_level} 个上级目录：{p_folder}')

        result = target_folder / p_folder / self.source.name
        return result.with_suffix(t_suffix)

    def iter_target_folders(self):
        for o in self.outputs:
            t = o.filename
            yield t.absolute().parent

    def __rich_repr__(self):
        yield 'from', self.source
        yield 'to', self.target
        yield 'general', self.general_options
        for x in self.inputs:
            yield 'input', x
        for x in self.outputs:
            yield 'output', x

    def check_output_writable(self):
        """检测全部目标文件是否可写入"""
        if self.profile.general.overwrite_existed:
            return True
        for o in self.outputs:
            file = o.filename
            if file.exists():
                return False
        return True


class MissionMaker:
    def __init__(self, profile):
        self.profile = profile

    def __call__(self, sources: [Path]):
        for s in sources:
            m = Mission(Path(s).absolute(), self.profile)
            env.debug(f'构建任务:', m)
            yield m



from collections import defaultdict
from pathlib import Path

from cx_core.utils import unquote_text


class OptionPackage:

    @staticmethod
    def __format_key(k: str):
        key = str(k).strip()
        if key.startswith('-'):
            key = key[1:]
        return key

    @staticmethod
    def __format_value(v):
        value = str(v) if v else ''
        return unquote_text(value)

    def __init__(self, filename=None):
        self.raw_data = defaultdict(list)
        self.filename = Path(filename) if filename else None

    def insert(self, key, value=None):
        k = OptionPackage.__format_key(key)
        v = OptionPackage.__format_value(value) if value else None
        if v:
            self.raw_data[k].append(v)
        elif k not in self.raw_data:
            self.raw_data[k] = []
        return self

    def iter_options(self):
        for k, vs in self.raw_data.items():
            for v in vs:
                value = v or ''
                yield k, value

    def iter_arguments(self):
        for k, v in self.iter_options():
            yield '-' + k
            yield v

    def __rich_repr__(self):
        yield 'filename', self.filename
        yield 'options', self.raw_data



import pkgutil
from argparse import ArgumentParser, ArgumentError
from pathlib import Path

from rich.columns import Columns
from rich.markdown import Markdown
from rich.panel import Panel
from rich_argparse import RichHelpFormatter

from cx_core import AbstractApp
from cx_core import DataPackage
from cx_core.app_logger import LogLevel
from cx_core.tui_elements import JobCounter
from cx_core.utils import normalize_path
from .env import env
from .mission import MissionMaker
from .planner import Planner
from .profile_loader import ProfileLoader, ProfileNoFoundError
from .script_writer import ScriptWriter
from .transcoder import Transcoder


class MediaKillerApp(AbstractApp):
    APP_VERSION = '0.2.4'
    APP_NAME = 'mediakiller'

    def __init__(self):
        super(MediaKillerApp, self).__init__()
        self.global_task = None

        parser = ArgumentParser(prog=MediaKillerApp.APP_NAME, formatter_class=RichHelpFormatter,
                                description='基于配置文件的批量转码工具',
                                epilog=f'Version {MediaKillerApp.APP_VERSION} Designed by xiii_1991')
        parser.add_argument('profile', help='指定配置文件路径', default=None, nargs='?')
        parser.add_argument('-g', '--generate-example-profile',
                            action="store_true", dest='generate_example',
                            help='生成范例文件')
        parser.add_argument('-a', '--add-source',
                            action='append', dest='sources', metavar='SOURCE_FILE',
                            help='增加来源文件')
        parser.add_argument('-s', '--make-script',
                            dest='script_output', metavar='SCRIPT_OUTPUT',
                            help='生成对应的脚本文件')
        parser.add_argument('-d', '--debug',
                            action='store_true', dest='debug', help='显示调试信息')
        parser.add_argument('--pretend', '-p',
                            dest='pretend_mode', action='store_true', help='空转模式，不执行命令')
        parser.add_argument('--full-help', help='显示详细的说明',
                            dest='full_help', action='store_true')
        parser.add_argument('-v', '--version', action='version', version=MediaKillerApp.APP_VERSION,
                            help='显示软件版本信息')
        self.profile = None
        self._parser = parser
        self.args = None

    @staticmethod
    def show_full_help():
        """显示完整的说明文件"""
        data = pkgutil.get_data('media_killer', 'help.md').decode('utf_8')
        panel = Panel(Markdown(data), width=80)
        env.console.print(panel)

    @staticmethod
    def copy_example_profile(tgt):
        tgt = normalize_path(tgt)
        if tgt.suffix != '.toml':
            tgt = tgt.with_suffix('.toml')
        data = pkgutil.get_data('media_killer', 'example_project.toml')
        try:
            with open(tgt, 'xb') as file:
                file.write(data)
            env.print(f'模板配置文件已保存到 {env.args.profile} ，[red]请在修改后运行！[/red]')
        except FileExistsError:
            env.error('文件 {0} 已存在，[red]请手动删除它，或指定其它的目标文件[/red]'.format(tgt))

    def __enter__(self):
        env.start()
        env.debug('env 环境已启动')

        _args = self._parser.parse_args()
        self.args = DataPackage(**vars(_args))
        env.debug('解析命令行参数：', self.args)

        env.log_level = LogLevel.DEBUG if self.args.debug else LogLevel.WARNING
        env.args = self.args

        self.global_task = env.progress.add_task(description='全局进度', start=False, visible=False)
        env.debug('构建全局进度条')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.stop_task(self.global_task)
        env.progress.remove_task(self.global_task)
        self.global_task = None
        env.debug('删除全局进度条')

        result = False
        if exc_type is None:
            pass
        elif issubclass(exc_type, ProfileNoFoundError):
            env.critical(exc_val)
            result = True
        elif issubclass(exc_type, ArgumentError):
            env.error('参数输入有误:', exc_val)
            result = True

        env.stop()
        return result

    def run(self):
        if env.args.full_help:
            env.info('检测到 full_help 设置，打印帮助文件并退出')
            MediaKillerApp.show_full_help()
            return

        if not env.args.profile:
            env.info('未检测到项目文件选项')
            raise ProfileNoFoundError

        if env.args.generate_example:
            env.info('检测到 "--generate-example" 参数，将拷贝配置文件模板')
            env.progress.update(task_id=self.global_task, description='正在输出配置文件', visible=True)
            self.copy_example_profile(env.args.profile)
            return

        with ProfileLoader(env.args.profile) as profile_loader:
            datas = profile_loader.load()
            self.profile = datas
        env.info('配置信息已初始化')

        if not self.profile.source.files:
            env.error('[red]未指定任何来源信息，无事可做[/red]')
            return

        planner = Planner(self.profile)
        env.progress.update(task_id=self.global_task, description='正在将来源信息加入计划', visible=True)
        planner.load_sources(self.profile.source.files)
        env.info('已发现 {0} 项来源'.format(len(planner)))

        env.progress.update(task_id=self.global_task, description='正在整理来源信息', visible=True)
        planner.arrange()
        env.print(f'整理结束，已添加 {len(planner)} 项计划')
        if env.args.debug:
            col = Columns(
                (Panel(Path(str(x)).name, style='yellow', border_style='green', safe_box=True,
                       title=f'#{i}', title_align='left')
                 for i, x in enumerate(planner)),
                expand=True, equal=False, column_first=True
            )
            env.debug(col)

        env.debug('构建 mission maker')
        missions = MissionMaker(self.profile)

        if env.args.script_output is not None:
            env.progress.update(task_id=self.global_task, description='输出脚本文件', visible=True)
            env.info('检测到 "--make-script" 参数，将进行脚本输出')
            output_file = normalize_path(env.args.script_output)
            with ScriptWriter(output_file) as writer:
                for mission in env.progress.track(missions(planner.plans), description=env.args.script_output):
                    writer.write_mission(mission)
            env.print(f'已输出脚本文件 "{env.args.script_output}"')
            return

        env.progress.update(self.global_task, description='总体进度', visible=True, total=len(planner))
        env.progress.start_task(self.global_task)

        job_counter = JobCounter(len(planner))
        for mission in missions(planner):
            job_counter.advance()
            if env.wanna_quit:
                env.print('正在取消未完成的任务...')
                break
            if env.args.pretend_mode:
                env.print(f'模拟运行: [yellow]{mission.source.name}[/yellow] 完毕')
            else:
                with Transcoder(mission) as coder:
                    coder.run()
            env.progress.advance(self.global_task)
            env.print(f'[yellow]{job_counter}[/yellow] [cyan]{mission.source.name}[/cyan] 转换完成')


def run():
    with MediaKillerApp() as media_killer:
        media_killer.run()



import signal

from rich.progress import Progress, SpinnerColumn

from cx_core import AbstractEnv, AppLogger, LogLevel


class Env(AbstractEnv):
    def __init__(self):
        super(Env, self).__init__()
        self.wanna_quit = False

        self._progress = Progress(
            SpinnerColumn(),
            *Progress.get_default_columns(),
            expand=True,
            transient=True)

        self._logger = AppLogger(console=self._progress.console)
        self._logger.level = LogLevel.WARNING

    def start(self):
        self._progress.start()
        self.info('全局环境已启动')

    def stop(self):
        self._progress.stop()
        self.info('Bye~')

    @property
    def console(self):
        return self._progress.console

    @property
    def progress(self):
        return self._progress


env = Env()


def signal_handler(sig, frame):
    if sig != signal.SIGINT:
        return
    env.info('接收到 SIGINT')
    env.print("收到终止信号，准备退出...")
    env.wanna_quit = True


signal.signal(signal.SIGINT, signal_handler)
